@@ -536,7 +536,7 @@ class Interp(VirtualMachine):
 
     def run_connect(self, jac_ast):
         """
-        connect: logical ( (NOT)? edge_ref expression)?;
+        connect: logical ( (NOT edge_ref | connect_op) expression)?;
         """
         kid = self.set_cur_ast(jac_ast)
         if len(kid) < 2:
@@ -561,12 +561,12 @@ class Interp(VirtualMachine):
             direction = kid[1].kid[0].name
             for i in target.obj_list():
                 for j in base.obj_list():
-                    use_edge = self.run_edge_ref(kid[1], is_spawn=True)
+                    use_edge = self.run_connect_op(kid[1])
                     self.rt_check_type(i, Node, kid[-1])
                     self.rt_check_type(j, Node, kid[-1])
-                    if direction == "edge_from":
+                    if direction == "connect_from":
                         j.attach_inbound(i, [use_edge])
-                    elif direction == "edge_to":
+                    elif direction == "connect_to":
                         j.attach_outbound(i, [use_edge])
                     else:
                         j.attach_bidirected(i, [use_edge])
@@ -1270,44 +1270,54 @@ class Interp(VirtualMachine):
         self.rt_error(f"Call to {str_op} is invalid.", jac_ast)
         return atom_res
 
-    def run_node_edge_ref(self, jac_ast):
+    def run_node_edge_ref(self, jac_ast, viable_nodes=None):
         """
         node_edge_ref:
-            node_ref filter_ctx?
-            | edge_ref (node_ref filter_ctx?)?;
+            node_ref filter_ctx? node_edge_ref?
+            | edge_ref node_edge_ref?;
         """
         kid = self.set_cur_ast(jac_ast)
+        result = JacSet()
         if kid[0].name == "node_ref":
-            result = self.run_node_ref(kid[0])
-            if len(kid) > 1:
-                result = self.run_filter_ctx(kid[1], result)
-            return JacValue(self, value=result)
+            result = self.run_node_ref(kid[0], viable_nodes=viable_nodes)
+            kid = kid[1:]
+            if len(kid) and kid[0].name == "filter_ctx":
+                result = self.run_filter_ctx(kid[0], result)
+                kid = kid[1:]
+            if len(kid):
+                result = self.run_node_edge_ref(kid[0], viable_nodes=result).value
 
         elif kid[0].name == "edge_ref":
-            relevant_edges = self.run_edge_ref(kid[0])
-            result = self.edge_to_node_jac_set(self.run_edge_ref(kid[0]))
-            if len(kid) > 1 and kid[1].name == "node_ref":
-                nres = self.run_node_ref(kid[1])
-                if len(kid) > 2:
-                    nres = self.run_filter_ctx(kid[2], nres)
-                result = result * nres
-                relevant_edges = self.edges_filter_on_nodes(relevant_edges, result)
-            self._relevant_edges = relevant_edges
-            return JacValue(self, value=result)
+            if not viable_nodes:
+                viable_nodes = [None]
+            self._relevant_edges = JacSet()
+            for i in viable_nodes:
+                relevant_edges = self.run_edge_ref(kid[0], location=i)
+                result += self.visibility_prune(
+                    self.edge_to_node_jac_set(relevant_edges, location=i)
+                )
+                kid = kid[1:]
+                if len(kid):
+                    result = self.run_node_edge_ref(kid[0], viable_nodes=result).value
 
-    def run_node_ref(self, jac_ast, is_spawn=False):
+                    relevant_edges = self.edges_filter_on_nodes(relevant_edges, result)
+                self._relevant_edges += relevant_edges
+        return JacValue(self, value=result)
+
+    def run_node_ref(self, jac_ast, is_spawn=False, viable_nodes=None):
         """
         node_ref: NODE_DBL_COLON NAME;
         """
         kid = self.set_cur_ast(jac_ast)
         if not is_spawn:
             result = JacSet()
+            viable_nodes = self.visibility_prune(viable_nodes)
             if len(kid) > 1:
-                for i in self.viable_nodes().obj_list():
+                for i in viable_nodes.obj_list():
                     if i.get_architype().is_instance(kid[-1].token_text()):
                         result.add_obj(i)
             else:
-                result += self.viable_nodes()
+                result += viable_nodes
         else:
             result = self.parent().run_architype(
                 kid[-1].token_text(), kind="node", caller=self
@@ -1348,42 +1358,24 @@ class Interp(VirtualMachine):
         )
         return JacValue(self, value=obj)
 
-    def run_edge_ref(self, jac_ast, is_spawn=False):
+    def run_edge_ref(self, jac_ast, location=None):
         """
         edge_ref: edge_to | edge_from | edge_any;
         """
         kid = self.set_cur_ast(jac_ast)
-        if not is_spawn:
-            return self.run_rule(kid[0])
-        else:
-            if len(kid[0].kid) > 2:
-                result = self.parent().run_architype(
-                    kid[0].kid[2].token_text(), kind="edge", caller=self
-                )
-                if kid[0].kid[3].name == "spawn_ctx":
-                    self.run_spawn_ctx(kid[0].kid[3], result)
-                elif kid[0].kid[3].name == "filter_ctx":
-                    self.rt_error("Filtering not allowed here", kid[0].kid[3])
-            else:
-                result = Edge(
-                    m_id=self._m_id,
-                    h=self._h,
-                    kind="edge",
-                    name="generic",
-                )
-            return result
+        return self.run_rule(kid[0], location)
 
-    def run_edge_to(self, jac_ast):
+    def run_edge_to(self, jac_ast, location=None):
         """
         edge_to:
             '-->'
             | '-' ('[' NAME (spawn_ctx | filter_ctx)? ']')? '->';
         """
         kid = self.set_cur_ast(jac_ast)
+        if not location:
+            location = self.current_node
         result = JacSet()
-        for i in (
-            self.current_node.outbound_edges() + self.current_node.bidirected_edges()
-        ):
+        for i in location.outbound_edges() + location.bidirected_edges():
             if len(kid) > 2 and not i.get_architype().is_instance(kid[2].token_text()):
                 continue
             result.add_obj(i)
@@ -1393,17 +1385,17 @@ class Interp(VirtualMachine):
             self.rt_error("Assigning values not allowed here", kid[3])
         return result
 
-    def run_edge_from(self, jac_ast):
+    def run_edge_from(self, jac_ast, location=None):
         """
         edge_from:
             '<--'
             | '<-' ('[' NAME (spawn_ctx | filter_ctx)? ']')? '-';
         """
         kid = self.set_cur_ast(jac_ast)
+        if not location:
+            location = self.current_node
         result = JacSet()
-        for i in (
-            self.current_node.inbound_edges() + self.current_node.bidirected_edges()
-        ):
+        for i in location.inbound_edges() + location.bidirected_edges():
             if len(kid) > 2 and not i.get_architype().is_instance(kid[2].token_text()):
                 continue
             result.add_obj(i)
@@ -1413,7 +1405,7 @@ class Interp(VirtualMachine):
             self.rt_error("Assigning values not allowed here", kid[3])
         return result
 
-    def run_edge_any(self, jac_ast):
+    def run_edge_any(self, jac_ast, location=None):
         """
         edge_any:
             '<-->'
@@ -1421,8 +1413,10 @@ class Interp(VirtualMachine):
         NOTE: these do not use strict bidirected semantic but any edge
         """
         kid = self.set_cur_ast(jac_ast)
+        if not location:
+            location = self.current_node
         result = JacSet()
-        for i in self.current_node.attached_edges():
+        for i in location.attached_edges():
             if len(kid) > 2 and not i.get_architype().is_instance(kid[2].token_text()):
                 continue
             result.add_obj(i)
@@ -1432,6 +1426,26 @@ class Interp(VirtualMachine):
             self.rt_error("Assigning values not allowed here", kid[3])
         return result
 
+    def run_connect_op(self, jac_ast):
+        """
+        connect_op: connect_to | connect_from | connect_any;
+        """
+        kid = self.set_cur_ast(jac_ast)
+        if len(kid[0].kid) > 2:
+            result = self.parent().run_architype(
+                kid[0].kid[2].token_text(), kind="edge", caller=self
+            )
+            if kid[0].kid[3].name == "spawn_ctx":
+                self.run_spawn_ctx(kid[0].kid[3], result)
+        else:
+            result = Edge(
+                m_id=self._m_id,
+                h=self._h,
+                kind="edge",
+                name="generic",
+            )
+        return result
+
     def run_list_val(self, jac_ast):
         """
         list_val: LSQUARE expr_list? RSQUARE;
@@ -1527,16 +1541,16 @@ class Interp(VirtualMachine):
 
     def run_spawn_edge(self, jac_ast):
         """
-        spawn_edge: expression edge_ref;
+        spawn_edge: expression connect_op;
         """
         kid = self.set_cur_ast(jac_ast)
         self.run_expression(kid[0])
         loc = self.pop().value
         if isinstance(loc, JacSet):
-            edge_set = [self.run_edge_ref(kid[1], is_spawn=True) for _ in loc]
+            edge_set = [self.run_connect_op(kid[1]) for _ in loc]
             loc = loc.obj_list()
         else:
-            edge_set = self.run_edge_ref(kid[1], is_spawn=True)
+            edge_set = self.run_connect_op(kid[1])
         return {
             "location": loc,
             "use_edge": edge_set,
@@ -1557,9 +1571,9 @@ class Interp(VirtualMachine):
             sp = self.run_spawn_edge(kid[0])
             if isinstance(sp["location"], Node):
                 ret_node = self.run_node_ref(kid[1], is_spawn=True)
-                if sp["direction"] == "edge_from":
+                if sp["direction"] == "connect_from":
                     sp["location"].attach_inbound(ret_node, [sp["use_edge"]])
-                elif sp["direction"] == "edge_to":
+                elif sp["direction"] == "connect_to":
                     sp["location"].attach_outbound(ret_node, [sp["use_edge"]])
                 else:
                     sp["location"].attach_bidirected(ret_node, [sp["use_edge"]])
@@ -1571,9 +1585,9 @@ class Interp(VirtualMachine):
                 sp["location"] = sp["location"].obj_list()
                 for i in range(len(sp["location"])):
                     ret_node = self.run_node_ref(kid[1], is_spawn=True)
-                    if sp["direction"] == "edge_from":
+                    if sp["direction"] == "connect_from":
                         sp["location"][i].attach_inbound(ret_node, [sp["use_edge"][i]])
-                    elif sp["direction"] == "edge_to":
+                    elif sp["direction"] == "connect_to":
                         sp["location"][i].attach_outbound(ret_node, [sp["use_edge"][i]])
                     else:
                         sp["location"][i].attach_bidirected(
@@ -1598,9 +1612,9 @@ class Interp(VirtualMachine):
             sp = self.run_spawn_edge(kid[0])
             if isinstance(sp["location"], Node):
                 ret_node = self.run_graph_ref(kid[1])
-                if sp["direction"] == "edge_from":
+                if sp["direction"] == "connect_from":
                     sp["location"].attach_inbound(ret_node, [sp["use_edge"]])
-                elif sp["direction"] == "edge_to":
+                elif sp["direction"] == "connect_to":
                     sp["location"].attach_outbound(ret_node, [sp["use_edge"]])
                 else:
                     sp["location"].attach_bidirected(ret_node, [sp["use_edge"]])
@@ -1610,9 +1624,9 @@ class Interp(VirtualMachine):
                 sp["location"] = sp["location"].obj_list()
                 for i in range(len(sp["location"])):
                     ret_node = self.run_graph_ref(kid[1])
-                    if sp["direction"] == "edge_from":
+                    if sp["direction"] == "connect_from":
                         sp["location"][i].attach_inbound(ret_node, [sp["use_edge"][i]])
-                    elif sp["direction"] == "edge_to":
+                    elif sp["direction"] == "connect_to":
                         sp["location"][i].attach_outbound(ret_node, [sp["use_edge"][i]])
                     else:
                         sp["location"][i].attach_bidirected(
@@ -1780,6 +1794,16 @@ class Interp(VirtualMachine):
             self.rt_error(f"Internal Exception: {e}", m._cur_jac_ast)
         self.inherit_runtime_state(m)
 
+    def visibility_prune(self, node_set=None):
+        """Returns all nodes that shouldnt be ignored"""
+        ret = JacSet()
+        if node_set is None:
+            node_set = self.current_node.attached_nodes()
+        for i in node_set:
+            if i not in self.ignore_node_ids.obj_list():
+                ret.add_obj(i)
+        return ret
+
     def run_rule(self, jac_ast, *args):
         """Helper to run rule if exists in execution context"""
         try:
